/* Copyright 2011 Frank Mitchell - Wildebeest License */
/* cddev.c - Spiegel v2.0 - Linux Kernel 2.6.16+ */
/* Contact: mitchell@wyatt672earp.force9.co.uk */

#include "spiegel.h"

void DRIVEEVAL(void) {
	const char *Here="@DRIVEEVAL";
	uint8_t Jbyte;
	uint16_t Jshort,BufCache;
	Jshort=sizeof(GTransBuf);
/* Mode Sense CD Capabilities & Mechanical Status Page */
	memset(&GRequest,0,sizeof(GRequest));
	GRequest.Cmd[0]=0x5A; /* MODE SENSE */
	GRequest.Cmd[1]=8; /* Disable Block Descriptors */
	GRequest.Cmd[2]=0x2A; /* Current CD Capabilities Page */
	BIGEND2(&GRequest.Cmd[7],&Jshort);
	GRequest.XferP=GTransBuf;
	GRequest.XferLen=sizeof(GTransBuf);
	GRequest.XferDirn=SG_DXFER_FROM_DEV;
	if(DOPACKET(GDevFil,"MODE SENSE CD Capabilities",&GRequest)!=0)goto HARDERR;
	SHOWCAPABILITIES(&BufCache);
	GDevCache=BufCache<<10; /* Device Cache Size Bytes */
	fprintf(SpiegLogF,"\nCDRW Hardware Buffer: %u Bytes\n",GDevCache);
/* Mode Sense Write Parameters Mode Page */
	memset(&GRequest,0,sizeof(GRequest));
	GRequest.Cmd[0]=0x5A; /* MODE SENSE */
	GRequest.Cmd[1]=8; /* Disable Block Descriptors */
	GRequest.Cmd[2]=0x05; /* Current Write Params Mode Page */
	BIGEND2(&GRequest.Cmd[7],&Jshort);
	GRequest.XferP=GTransBuf;
	GRequest.XferLen=sizeof(GTransBuf);
	GRequest.XferDirn=SG_DXFER_FROM_DEV;
	if(DOPACKET(GDevFil,"MODE SENSE Current Write Parameters",&GRequest)!=0)goto HARDERR;
	SHOWWRITEPARAMS("Initial Drive Settings");
/* Write Page Defaults Clear Firmware Data Using Zero Test Write Bit */
	DEFLTWRITE();
/* Get Event/Status Notification */
	fprintf(SpiegLogF,"\nGET EVENT/STATUS NOTIFICATION\n");
	memset(&GRequest,0,sizeof(GRequest));
	GRequest.Cmd[0]=0x4A; /* GET EVENT/STATUS NOTIFICATION */
	GRequest.Cmd[1]=1; /* Immed/Polled */
	GRequest.Cmd[4]=0x54; /* Busy + Media + Power Classes */
	BIGEND2(&GRequest.Cmd[7],&Jshort);
	GRequest.XferP=GTransBuf;
	GRequest.XferLen=sizeof(GTransBuf);
	GRequest.XferDirn=SG_DXFER_FROM_DEV;
	if(DOPACKET(GDevFil,"GET EVENT/STATUS NOTIFICATION",&GRequest)!=0)goto BADNOTIFGET;
	Jbyte=GTransBuf[2]&0x07;
	fprintf(SpiegLogF,"Notification Returned: %u, %s\n",Jbyte,NOTIFICATIONS(Jbyte));
	if(!(GTransBuf[3]&0x40))goto NOBUSYPOLL;
	return;
HARDERR:printf("Failed Hardware Request %s\n",Here);goto STOP;
BADNOTIFGET:printf("Failed Event/Status Notification Request %s\n",Here);goto STOP;
NOBUSYPOLL:printf("Cannot Poll For Device Busy Events %s\n",Here);goto STOP;
STOP:printf("errno Says: %s\n",strerror(errno));exit(EXIT_FAILURE);
} /* Hardware Parameter Pages */

void SHOWCAPABILITIES(uint16_t *BufCacheP) {
	const char *Here="@SHOWCAPABILITIES";
	uint8_t Jbyte,CapsFound;
	char *RotCtl;
	uint16_t Jshort,StartByte,WriteSpeed;
	int j;
	fprintf(SpiegLogF,"\nCAPABILITIES & MECHANICAL STATUS\n");
/* Header Adds Eight Bytes To Mode Page Locations */
	j=(uint8_t)GTransBuf[9]+10;
	memset(&GTransBuf[j],0,2048-j);
	fprintf(SpiegLogF,"Reads CD-R: %c\n",(GTransBuf[10]&1)?'Y':'N');
	fprintf(SpiegLogF,"Reads CD-RW: %c\n",(GTransBuf[10]&2)?'Y':'N');
	fprintf(SpiegLogF,"Reads Method 2 Fixed Packet Addressing: %c\n",(GTransBuf[10]&4)?'Y':'N');
	fprintf(SpiegLogF,"Reads DVD-ROM: %c\n",(GTransBuf[10]&8)?'Y':'N');
	fprintf(SpiegLogF,"Reads DVD-R: %c\n",(GTransBuf[10]&16)?'Y':'N');
	fprintf(SpiegLogF,"Reads DVD-RAM: %c\n",(GTransBuf[10]&32)?'Y':'N');
	fprintf(SpiegLogF,"Writes CD-R: %c\n",(GTransBuf[11]&1)?'Y':'N');
	fprintf(SpiegLogF,"Writes CD-RW: %c\n",(GTransBuf[11]&2)?'Y':'N');
	fprintf(SpiegLogF,"Test Write Function Supported: %c\n",(GTransBuf[11]&4)?'Y':'N');
	CapsFound=((uint8_t)GTransBuf[11]&7); /* CD-R/RW Write & Test Capability */
	fprintf(SpiegLogF,"Writes DVD-R: %c\n",(GTransBuf[11]&16)?'Y':'N');
	fprintf(SpiegLogF,"Writes DVD-RAM: %c\n",(GTransBuf[11]&32)?'Y':'N');
	fprintf(SpiegLogF,"Reads Mode 2 Form 1 CD Sectors: %c\n",(GTransBuf[12]&16)?'Y':'N');
	fprintf(SpiegLogF,"Reads Mode 2 Form 2 CD Sectors: %c\n",(GTransBuf[12]&32)?'Y':'N');
	fprintf(SpiegLogF,"Reads Multi-Session CD Disks: %c\n",(GTransBuf[12]&64)?'Y':'N');
	fprintf(SpiegLogF,"Supports Buffer Under-Run Free CDR/RW Recording: %c\n",(GTransBuf[12]&128)?'Y':'N');
	CapsFound|=((uint8_t)GTransBuf[12]&0xC0); /* Under-Run Prot & Multi-Session Capability */
	fprintf(SpiegLogF,"Reads Raw R-W Sub-Channel From Lead-In: %c\n",(GTransBuf[15]&32)?'Y':'N');
	BIGEND2(&Jshort,GTransBuf+16);
	fprintf(SpiegLogF,"Maximum Read Speed (MMC-1) KB/sec: %hu\n",Jshort);
	BIGEND2(BufCacheP,GTransBuf+20);
	fprintf(SpiegLogF,"Buffer Cache Size KB: %hu\n",*BufCacheP);
	BIGEND2(&Jshort,GTransBuf+22);
	fprintf(SpiegLogF,"Current Read Speed (MMC-1) KB/sec: %hu\n",Jshort);
	BIGEND2(&Jshort,GTransBuf+26);
	fprintf(SpiegLogF,"Maximum Write Speed (MMC-1) KB/sec: %hu\n",Jshort);
	BIGEND2(&Jshort,GTransBuf+28);
	fprintf(SpiegLogF,"Current Write Speed (MMC-1) KB/sec: %hu\n",Jshort);
	fprintf(SpiegLogF,"Pure CAV Rotation Control Selected: %c\n",(GTransBuf[35]&0x03)?'Y':'N');
	BIGEND2(&Jshort,GTransBuf+36);
	fprintf(SpiegLogF,"Current Write Speed (MMC-3) KB/sec: %hu\n",Jshort);
	BIGEND2(&Jshort,GTransBuf+38);
	fprintf(SpiegLogF,"Number Of Write Speed Performance Descriptors: %hu\n",Jshort);
	for(j=1;j<=Jshort;j++) {
		StartByte=40+((j-1)<<2);
		Jbyte=(uint8_t)GTransBuf[StartByte+1]&7;
		switch(Jbyte) {
			case 0:RotCtl="CLV/Part-CAV/Default";break;
			case 1:RotCtl="Pure CAV";break;
			default:RotCtl="Rotation Control Undefined";
		} /* Rotation Control Types */
		BIGEND2(&WriteSpeed,GTransBuf+StartByte+2);
		fprintf(SpiegLogF,"Write Speed KB/sec: %hu, %s\n",WriteSpeed,RotCtl);
	} /* List Write Speed Performance Descriptors */
	if(CapsFound<0xC5)goto NOCAP;
	return;
NOCAP:printf("Drive Lacks Essential Capability %s\n",Here);goto STOP;
STOP:printf("errno Says: %s\n",strerror(errno));exit(EXIT_FAILURE);
} /* CD Capabilities & Mechanical Status Page */

void SHOWWRITEPARAMS(const char *Desc) {
	uint8_t Jbyte;
	uint32_t Jlong;
	int j;
	fprintf(SpiegLogF,"\nCURRENT WRITE PARAMETERS: %s\n",Desc);
/* Header Adds Eight Bytes To Mode Page Locations */
	j=(uint8_t)GTransBuf[9]+10;
	memset(&GTransBuf[j],0,2048-j);
	Jbyte=(uint8_t)GTransBuf[10]&0x0F;
	fprintf(SpiegLogF,"Write Stream Type: %02Xh, %s\n",Jbyte,WRITETYPES(Jbyte));
	fprintf(SpiegLogF,"Test Writing Enabled: %c\n",(GTransBuf[10]&16)?'Y':'N');
	fprintf(SpiegLogF,"Link Size Valid: %c\n",(GTransBuf[10]&32)?'Y':'N');
	fprintf(SpiegLogF,"Buffer Underrun Protection Enabled: %c\n",(GTransBuf[10]&64)?'Y':'N');
	Jbyte=(uint8_t)GTransBuf[11]&0x0F; /* Bit 1 Means Copy Permitted */
	fprintf(SpiegLogF,"Track Mode Control Nibble: %02Xh, %s\n",Jbyte,TRACKMODES(Jbyte));
	fprintf(SpiegLogF,"Fixed Packet Type: %c\n",(GTransBuf[11]&32)?'Y':'N');
	Jbyte=((uint8_t)GTransBuf[11]&0xC0)>>6;
	fprintf(SpiegLogF,"Multisession Closure: %u, %s\n",Jbyte,MULTISESSCLOSES(Jbyte));
	Jbyte=(uint8_t)GTransBuf[12]&0x0F;
	fprintf(SpiegLogF,"Data Block Type: %u, %s\n",Jbyte,DATABLOCKTYPES(Jbyte));
	fprintf(SpiegLogF,"Linking Loss Area Size: %u\n",(uint8_t)GTransBuf[13]);
	Jbyte=(uint8_t)GTransBuf[16];
	fprintf(SpiegLogF,"Session Format: %02Xh, %s\n",Jbyte,SESSIONFORMATS(Jbyte));
	BIGEND4(&Jlong,GTransBuf+18);
	fprintf(SpiegLogF,"Fixed Packet Size, Blocks: %d\n",Jlong);
	fprintf(SpiegLogF,"Sub-Header Hex:");
	for(j=56;j<60;j++)fprintf(SpiegLogF," %02X",(uint8_t)GTransBuf[j]);
	fprintf(SpiegLogF,"\n");
	return;
} /* Write Parameters Mode Page */

const char *WRITETYPES(uint8_t Code) {
	switch(Code) {
		case 0x00:
		return "Packet/Incremental";
		case 0x01:
		return "Track-At-Once";
		case 0x02:
		return "Session-At-Once";
		case 0x03:
		return "Raw";
		default:
		return "Reserved Write Type";
	}
} /* Write Types */

const char *TRACKMODES(uint8_t Code) {
	switch(Code) {
		case 0x00:
		return "Two Audio Channels No Pre-Emphasis";
		case 0x01:
		return "Two Audio Channels With Pre-Emphasis";
		case 0x02:
		return "Two Audio Channels No Pre-Emphasis";
		case 0x03:
		return "Two Audio Channels With Pre-Emphasis";
		case 0x04:
		return "Data Track Uninterrupted";
		case 0x05:
		return "Data Track Incremental";
		case 0x06:
		return "Data Track Uninterrupted";
		case 0x07:
		return "Data Track Incremental";
		case 0x08:
		return "Four Audio Channels No Pre-Emphasis";
		case 0x09:
		return "Four Audio Channels With Pre-Emphasis";
		case 0x0A:
		return "Four Audio Channels No Pre-Emphasis";
		case 0x0B:
		return "Four Audio Channels With Pre-Emphasis";
		default:
		return "Reserved Track Mode";
	}
} /* Track Modes */

const char *MULTISESSCLOSES(uint8_t Code) {
	switch(Code) {
		case 0:
		return "No B0 Pointer, Next Session Not Allowed";
		case 1:
		return "B0 Pointer = FF:FF:FF, Next Session Not Allowed";
		case 3:
		return "Next Session Allowed, B0 Pointer = Next Program Area";
		default:
		return "Reserved Multisession Field";
	}
} /* Multisession Closures */

const char *DATABLOCKTYPES(uint8_t Code) {
	switch(Code) {
		case 0:
		return "Raw Data *2352, Not Valid For Packet";
		case 1:
		return "Raw Data *2368 With P&Q Sub-Channel";
		case 2:
		return "Raw Data *2448 With P-W Sub-Channel";
		case 3:
		return "Raw Data *2448 With Raw P-W Sub-Channel";
		case 7:
		return "Vendor Specific Data Block Type";
		case 8:
		return "Mode 1 *2048 ISO/IEC 10149 (Mandatory)";
		case 9:
		return "Mode 2 *2336 ISO/IEC 10149";
		case 10:
		return "Mode 2 *2048 CD-ROM XA, Form 1 (Mandatory)";
		case 11:
		return "Mode 2 *2056 CD-ROM XA, Form 1, With Sub-header";
		case 12:
		return "Mode 2 *2324 CD-ROM XA, Form 2";
		case 13:
		return "Mode 2 *2332 CD-ROM XA, Form 1, Form 2, Mixed (Mandatory)";
		case 15:
		return "Vendor Specific Data Block Type";
		default:
		return "Reserved Data Block Type";
	}
} /* Data Block Types */

const char *SESSIONFORMATS(uint8_t Code) {
	switch(Code) {
		case 0x00:
		return "CD-DA Or CD-ROM Type Data Disc";
		case 0x10:
		return "CD-I Disc";
		case 0x20:
		return "CD-ROM XA Or DDCD Disc";
		default:
		return "Reserved Session Format Code";
	}
} /* Session Format Types */

int DISCEVAL(uint16_t *FirstTrackP,uint16_t *LastTrackP,uint16_t *SessTotP) {
	const char *Here="@DISCEVAL";
	int DiscStatus;
	uint16_t Jshort;
	uint32_t Capacity,BlockLen;
	float MegUsed;
/* Read CD Recorded Capacity */
	memset(&GRequest,0,sizeof(GRequest));
	GRequest.Cmd[0]=0x25; /* READ CD RECORDED CAPACITY */
	GRequest.XferP=GTransBuf;
	GRequest.XferLen=sizeof(GTransBuf);
	GRequest.XferDirn=SG_DXFER_FROM_DEV;
	if(DOPACKET(GDevFil,"READ CAPACITY",&GRequest)!=0)goto MEDIAERR;
	fprintf(SpiegLogF,"\nREAD CAPACITY\n");
	BIGEND4(&Capacity,&GTransBuf[0]);
	fprintf(SpiegLogF,"Last Valid Seek LBA: %d\n",Capacity);
	GSeekLBA=Capacity;
	BIGEND4(&BlockLen,&GTransBuf[4]);
	fprintf(SpiegLogF,"Block Length: %d\n",BlockLen);
	if(BlockLen!=2048)goto BLOCKWRONG;
	MegUsed=(float)Capacity/512.0;
	printf("Capacity Already Used: %3.1f Meg\n",MegUsed);
/* Read Disc Information */
	memset(&GRequest,0,sizeof(GRequest));
	GRequest.Cmd[0]=0x51; /* READ DISC INFORMATION */
	Jshort=34;
	BIGEND2(&GRequest.Cmd[7],&Jshort);
	GRequest.XferP=GTransBuf;
	GRequest.XferLen=sizeof(GTransBuf);
	GRequest.XferDirn=SG_DXFER_FROM_DEV;
	if(DOPACKET(GDevFil,"READ DISC INFORMATION",&GRequest)!=0)goto MEDIAERR;
	DiscStatus=DISCINFO(FirstTrackP,LastTrackP,SessTotP);
	return DiscStatus;
BLOCKWRONG:printf("Wrong Block Length: 2048 Needed On Data CDs %s\n",Here);goto STOP;
MEDIAERR:printf("Failed Media Request %s\n",Here);goto STOP;
STOP:printf("errno Says: %s\n",strerror(errno));exit(EXIT_FAILURE);
} /* Disc Evaluation */

int DISCINFO(uint16_t *FirstTrackP,uint16_t *LastTrackP,uint16_t *SessTotP) {
	const char *Here="@DISCINFO";
	uint8_t Jbyte;
	uint16_t Jshort,SessTrackN;
	uint32_t Jlong;
	int DiscType,LastSessState,DiscStatus;
	fprintf(SpiegLogF,"\nREAD DISC INFORMATION\n");
	BIGEND2(&Jshort,GTransBuf);
	memset(&GTransBuf[Jshort+2],0,2046-Jshort);
	Jbyte=(uint8_t)GTransBuf[2]&0x03;
	fprintf(SpiegLogF,"Disc Status: %u, %s\n",Jbyte,DISCSTATUSS(Jbyte));
	DiscStatus=Jbyte;
	Jbyte=((uint8_t)GTransBuf[2]&0x0C)>>2;
	fprintf(SpiegLogF,"Last Session State: %u, %s\n",Jbyte,LASTSESSIONS(Jbyte));
	LastSessState=Jbyte;
	fprintf(SpiegLogF,"Erasable RW Medium: %c\n",(GTransBuf[2]&16)?'Y':'N');
	fprintf(SpiegLogF,"First Track Number: %u\n",(uint8_t)GTransBuf[3]);
	*FirstTrackP=(uint16_t)GTransBuf[3];
	SPLITEND2(&Jshort,&GTransBuf[9],&GTransBuf[4]);
	fprintf(SpiegLogF,"Number Of Sessions: %hu\n",Jshort);
	*SessTotP=Jshort;
	SPLITEND2(&Jshort,&GTransBuf[10],&GTransBuf[5]);
	fprintf(SpiegLogF,"First Track In Last Session: %hu\n",Jshort);
	SessTrackN=Jshort;
	SPLITEND2(&Jshort,&GTransBuf[11],&GTransBuf[6]);
	fprintf(SpiegLogF,"Last Track In Last Session: %hu\n",Jshort);
	*LastTrackP=Jshort;
	Jbyte=(uint8_t)(GTransBuf[8]);
	fprintf(SpiegLogF,"Disc Type: %02Xh, %s\n",Jbyte,DISCTYPES(Jbyte));
	DiscType=(uint8_t)GTransBuf[8];
	BIGEND4(&Jlong,GTransBuf+12);
	fprintf(SpiegLogF,"Disc Identification: %d\n",Jlong);
	fprintf(SpiegLogF,"Number Of Power-Calibrated Speeds: %u\n",(uint8_t)GTransBuf[33]);
/* Cannot Continue With Inappropriate Disc */
	if(DiscStatus>1)goto NOAPPEND;
	if(LastSessState!=0)goto NONEMPTYSESS;
	return DiscStatus;
NOAPPEND:printf("Disc Status Not Empty Or Appendable %s\n",Here);goto STOP;
NONEMPTYSESS:printf("Last Session On Disc Not Empty %s\n",Here);goto STOP;
STOP:printf("errno Says: %s\n",strerror(errno));exit(EXIT_FAILURE);
} /* CDB 0x51, READ DISC INFORMATION */

const char *DISCSTATUSS(uint8_t Code) {
	switch(Code) {
		case 0:
		return "Empty Disc";
		case 1:
		return "Incomplete Disc, Appendable Session";
		case 2:
		return "Finalized, Last Session Closed";
		case 3:
		return "Random Access Only, No Sessions";
		default:
		return "Reserved Disc Status Value";
	}
} /* Disc Statuses */

const char *LASTSESSIONS(uint8_t Code) {
	switch(Code) {
		case 0:
		return "Empty Session";
		case 1:
		return "Incomplete Session";
		case 2:
		return "Damaged Session";
		case 3:
		return "Complete Session";
		default:
		return "Reserved Session State Value";
	}
} /* Last Session States */

const char *DISCTYPES(uint8_t Code) {
	switch(Code) {
		case 0x00:
		return "CD-DA or CD-ROM Disc";
		case 0x10:
		return "CD-I Disc";
		case 0x20:
		return "CD-ROM XA Disc Or DDCD";
		case 0xFF:
		return "Undefined Disc Type";
		default:
		return "Reserved Disc Type";
	}
} /* Disc Types */

uint32_t SESSEVAL(uint32_t *WriteStartP,uint16_t FirstTrackN,uint16_t LastTrackN) {
	const char *Here="@SESSEVAL";
	uint16_t Jshort;
	uint32_t Jlong,FreeBlocks;
/* Read Track Information For All Tracks */
	for(Jlong=FirstTrackN;Jlong<=LastTrackN;Jlong++) {
		memset(&GRequest,0,sizeof(GRequest));
		GRequest.Cmd[0]=0x52; /* READ TRACK INFORMATION */
		GRequest.Cmd[1]=1; /* Specify Track Number */
		BIGEND4(&GRequest.Cmd[2],&Jlong);
		Jshort=40;
		BIGEND2(&GRequest.Cmd[7],&Jshort);
		GRequest.XferP=GTransBuf;
		GRequest.XferLen=sizeof(GTransBuf);
		GRequest.XferDirn=SG_DXFER_FROM_DEV;
		if(DOPACKET(GDevFil,"READ TRACK INFORMATION",&GRequest)!=0)goto MEDIAERR;
		FreeBlocks=TRACKINFO(WriteStartP);
	} /* Writable Disk Returns Data From Invisible Track */
	return FreeBlocks;
MEDIAERR:printf("Failed Media Request %s\n",Here);goto STOP;
STOP:printf("errno Says: %s\n",strerror(errno));exit(EXIT_FAILURE);
} /* Sessions Evaluation */

uint32_t TRACKINFO(uint32_t *WriteStartP) {
	uint8_t Jbyte;
	uint16_t Jshort;
	uint32_t Jlong,FreeBlocks;
	fprintf(SpiegLogF,"\nREAD TRACK INFORMATION\n");
	BIGEND2(&Jshort,GTransBuf);
	memset(&GTransBuf[Jshort+2],0,2046-Jshort);
	SPLITEND2(&Jshort,&GTransBuf[32],&GTransBuf[2]);
	fprintf(SpiegLogF,"Track Number: %hu\n",Jshort);
	SPLITEND2(&Jshort,&GTransBuf[33],&GTransBuf[3]);
	fprintf(SpiegLogF,"Session Number: %hu\n",Jshort);
	Jbyte=(uint8_t)GTransBuf[5]&0x0F; /* Bit 1 Means Copy Permitted */
	fprintf(SpiegLogF,"Track Mode Control Nibble: %02Xh, %s\n",Jbyte,TRACKMODES(Jbyte));
	Jbyte=(uint8_t)GTransBuf[6]&0x0F;
	fprintf(SpiegLogF,"Data Mode: %02Xh, %s\n",Jbyte,DATAMODES(Jbyte));
	fprintf(SpiegLogF,"Track Must Be Written With Packets: %c\n",(GTransBuf[6]&32)?'Y':'N');
	fprintf(SpiegLogF,"Track Is Blank: %c\n",(GTransBuf[6]&64)?'Y':'N');
	fprintf(SpiegLogF,"Track Is Reserved: %c\n",(GTransBuf[6]&128)?'Y':'N');
	fprintf(SpiegLogF,"Next Writable Address Valid: %c\n",(GTransBuf[7]&1)?'Y':'N');
	fprintf(SpiegLogF,"Last Recorded Address Valid: %c\n",(GTransBuf[7]&2)?'Y':'N');
	BIGEND4(&Jlong,GTransBuf+8);
	fprintf(SpiegLogF,"Track Start Address: %d\n",Jlong);
	BIGEND4(WriteStartP,GTransBuf+12);
	fprintf(SpiegLogF,"Next Writable Address: %d\n",*WriteStartP);
	BIGEND4(&FreeBlocks,GTransBuf+16);
	fprintf(SpiegLogF,"Free Blocks: %d\n",FreeBlocks);
	BIGEND4(&Jlong,GTransBuf+20);
	fprintf(SpiegLogF,"Fixed Packet Size: %d\n",Jlong);
	BIGEND4(&Jlong,GTransBuf+24);
	fprintf(SpiegLogF,"Track Size: %d\n",Jlong);
	BIGEND4(&Jlong,GTransBuf+28);
	fprintf(SpiegLogF,"Last Recorded Address: %d\n",Jlong);
	BIGEND4(&Jlong,GTransBuf+36);
	fprintf(SpiegLogF,"Read Compatibility LBA: %d\n",Jlong);
	return FreeBlocks;
} /* CDB 0x52, READ TRACK INFORMATION */

const char *DATAMODES(uint8_t Code) {
	switch(Code) {
		case 0x01:
		return "Mode 1 ISO/IEC 10149";
		case 0x02:
		return "Mode 2 ISO/IEC 10149, CD-ROM XA, DDCD";
		case 0x0F:
		return "Data Block Type Unknown, No Track Descriptor Block";
		default:
		return "Reserved Data Mode Type";
	}
} /* Data Modes */
