/* Copyright 2011 Frank Mitchell - Wildebeest License */
/* driver.c - Spiegel v2.0 - Linux Kernel 2.6.16+ */
/* Contact: mitchell@wyatt672earp.force9.co.uk */

#include "spiegel.h"

int main(int argc,char **argv) {
	const char *Here="@MAIN";
	uint8_t WriteMode,NextSess,CloseFunc;
	int Kyn,DiscStatus,IsoFildes;
	uint32_t WriteStart,WriteEnd,IsoBlocks,
	  NeedBlocks,FreeBlocks,CdSpeed,DvdSpeed,WritePerf;
	uint16_t FirstTrackN,LastTrackN,SessTot;
	float Frac,FreeMeg,NeedMeg;
	struct stat IsoStat;
	printf("Spiegel Uses Real-Time Priority: Be Root!\n");
/* Command Line Options & Log Files */
	OPTEVAL(argc,argv,&CdSpeed,&DvdSpeed);
	SYSEVAL(argc,argv); /* Check System Data */
	STARTUP(); /* Unit Startup */
	DRIVEEVAL(); /* Read Hardware Data Pages */
	WriteMode=CONFIGEVAL(); /* Check Profile List For Write Type Needed */
/* Overall Streaming Speeds */
	WriteStart=0; /* Negative Values Not Accepted */
	WriteEnd=0x7FFFFFFF;
	WritePerf=(WriteMode&0x01)?CdSpeed:DvdSpeed; /* CD Or DVD ? */
	WritePerf=WRITESPEED(WriteStart,WriteEnd,WritePerf,1,
	  "Initial Mixed Read/Write Streaming Speeds");
/* Read Disk Track & Session Data */
	DiscStatus=DISCEVAL(&FirstTrackN,&LastTrackN,&SessTot);
	FreeBlocks=SESSEVAL(&WriteStart,FirstTrackN,LastTrackN);
	FreeMeg=(float)FreeBlocks/512.0;
	printf("Before Run Free Space: %3.1f Meg\n",FreeMeg);
/* Can Set Symbolic Link Default: "ln -sf ImagePath.iso IsoLink" */
	GETSTR("\tChoose Iso Image Path [%s]\n",GIsoFilPath,"IsoLink",UFSPATHL);
	fprintf(SpiegLogF,"\nChosen Iso Image Path:\n\t%s\n",GIsoFilPath);
	if((IsoFildes=open(GIsoFilPath,O_RDONLY))==-1)goto BADOPENFIL;
	if(stat(GIsoFilPath,&IsoStat)==-1)goto COULDNTSTAT;
	GIsoSize=IsoStat.st_size;
	close(IsoFildes);
/* Check Enough Data Blocks Available */
	if(GIsoSize&2047)printf("WARNING: CD Image Ends With Partial Sector\n");
	IsoBlocks=(GIsoSize+2047)>>11;
	if((IsoBlocks>360000)&&(GIsoSize&32767))
	  printf("NOTE: DVD Image Ends With Partial Superblock\n");
	NeedBlocks=(IsoBlocks>MINTRACKL)?IsoBlocks:MINTRACKL;
	NeedBlocks=(WriteMode&0x01)?NeedBlocks:(((NeedBlocks+15)>>4)<<4); /* DVD Superblocks */
	NeedMeg=(float)NeedBlocks/512.0;
	Frac=NeedMeg*100.0/FreeMeg;
	printf("Space Needed: %3.1f Meg, Free Space Fraction: %4.2f%%\n",NeedMeg,Frac);
	FreeMeg=(float)((int)FreeBlocks-(int)NeedBlocks)/512.0; /* Blocks To Megabytes */
	printf("After Run Free Space < %3.1f Meg\n",FreeMeg);
	if(NeedBlocks>FreeBlocks)goto NOROOM;
	WriteEnd=WriteStart+NeedBlocks;
/* Specify Multi Session Mode */
	Kyn=GETYN("\tAllow Further Sessions? [%c]\n",'Y');
	if(Kyn=='Y') {
		NextSess=0xC6; /* Multi-Session Field=11b */
		CloseFunc=2; /* Normal CLOSE SESSION */
	} /* Leaves Disk Open */
	else {
		NextSess=0x06; /* Next Session B0 Pointer Disabled */
		CloseFunc=(WriteMode&0x08)?5:2; /* DVD+R CLOSE SESSION */
	} /* For Write-Protected Disk */
/* Write Cycle Loop */
	do {
		WritePerf=GETINT("\tSet Write Speed KB/Sec [%d]\n",WritePerf);
		WritePerf=WRITESPEED(WriteStart,WriteEnd,WritePerf,0,
		  "Independent Write Speed Set For Recorded Region");
		WriteMode=WRITEPARMS(WriteMode,NextSess);
		Kyn=GETYN("\tLast Chance To Abandon Run! Proceed? [%c]\n",'Y');
		if(Kyn=='N')break;
		Kyn=TESTBURN(WriteMode,WriteStart,LastTrackN,CloseFunc);
		if(Kyn)goto OFFLINE;
		DEFLTWRITE(); /* Clear Test Track Data */
		if((WriteMode&0x10)==0)break; /* Write Run Done */
		Kyn=GETYN("\tThat Was A Test: Repeat Run? [%c]\n",'Y');
		if(Kyn=='N')break;
	} while(1);
/* Restore Streaming Speed Defaults */
	WRITESPEED(0,0x7FFFFFFF,0x7FFFFFFF,4,"Restore Streaming Speed Defaults");
/* Stop Unit */
	memset(&GRequest,0,sizeof(GRequest));
	GRequest.Cmd[0]=0x1B; /* START STOP UNIT */
	GRequest.Cmd[4]=0; /* No-Unload Stop */
	GRequest.XferDirn=SG_DXFER_NONE;
	if(DOPACKET(GDevFil,"START STOP UNIT",&GRequest)!=0)goto BADSTOP;
OFFLINE:
	if(close(GDevFil)!=0)goto BADCLOSEDEV;
	fprintf(SpiegLogF,"\n%s\n%s\n",SAPPMES,CPRTMES);
/* Close Recording Log File */
	fclose(SpiegLogF);
	printf("REMOVE DISK NOW\n");
	exit(EXIT_SUCCESS);
BADOPENFIL:printf("Couldn't open() Iso File %s\n",Here);goto STOP;
COULDNTSTAT:printf("Couldn't stat() Iso File %s\n",Here);goto STOP;
NOROOM:printf("Insufficient User Blocks On Empty Track %s\n",Here);goto STOP;
BADSTOP:printf("Failed Unit Stop %s\n",Here);goto STOP;
BADCLOSEDEV:printf("Failed Device File Close %s\n",Here);goto STOP;
STOP:printf("errno Says: %s\n",strerror(errno));exit(EXIT_FAILURE);
} /* Driver */

void OPTEVAL(int argc,char **argv,uint32_t *CdSpeedP,uint32_t *DvdSpeedP) {
	const char *Here="@OPTEVAL";
	int Opt,Arg,PrioMin,PrioMax,PrioMid;
	struct sched_param RtParm;
	GBufFac=11; /* Default Write Buffer Ring Factor */
	*CdSpeedP=0x7FFFFFFF; /* Default CD Speed Limit */
	*DvdSpeedP=0x7FFFFFFF; /* Default DVD Speed Limit */
	PrioMin=sched_get_priority_min(SCHED_FIFO);
	PrioMax=sched_get_priority_max(SCHED_FIFO);
	PrioMid=(PrioMin+PrioMax)>>1;
	RtParm.sched_priority=PrioMid; /* Default Mid-Range Real-Time Priority */
	if(sched_setscheduler(0,SCHED_FIFO,&RtParm)==-1)goto NOREALTIME;
	while((Opt=getopt(argc,argv,"b:c:d:hp:v"))!=-1) {
		switch(Opt) {
		case 'b': /* Set Write Buffer Ring Factor */
			sscanf(optarg," %d",&Arg);
			GBufFac=(Arg<10)?10:((Arg>20)?20:Arg);
			break;
		case 'c': /* Set CD Speed Limit, KB/Sec */
/* 4x=706,8x=1411,10x=1764,16x=2822,24x=4234,32x=5645,40x=7056,48x=8467 */
			sscanf(optarg," %d",CdSpeedP);
			break;
		case 'd': /* Set DVD Speed Limit, KB/Sec */
/* 1x=1385,2x=2770,4x=5540,6x=8310,8x=11080,12x=16620,16x=22160 */
			sscanf(optarg," %d",DvdSpeedP);
			break;
		default:
		case 'h': /* Display Brief Help */
			printf("Spiegel Uses Real-Time Priority: Be Root!\n");
			printf("IsoLink = Default Symbolic Link To Image File\n");
/* Write Buffer Ring Size = Device Buffer * 1.25 [10/8] - 2.5 [20/8] */
			printf("-b INT: Set Write Buffer Ring Factor 10-20 [11]\n");
			printf("-c INT: Set CD Speed Limit [Max KB/Sec]\n");
			printf("-d INT: Set DVD Speed Limit [Max KB/Sec]\n");
			printf("-h    : Display Brief Help\n");
			printf("-p INT: Set Real-Time Priority %d-%d [%d]\n",
			  PrioMin,PrioMax,PrioMid);
			printf("-v    : Display Version & Licensing Message\n");
			exit(EXIT_SUCCESS);
		case 'p': /* Set Real-Time Priority */
			sscanf(optarg," %d",&Arg);
			RtParm.sched_priority=(Arg<PrioMin)?PrioMin:((Arg>PrioMax)?PrioMax:Arg);
			sched_setscheduler(0,SCHED_FIFO,&RtParm);
			break;
		case 'v': /* Display Version & Licensing Message */
			printf("\n"SAPPMES"\n"CPRTMES"\n");
			exit(EXIT_SUCCESS);
		} /* Process Options */
	} /* Scan Command Line */
	if((SpiegLogF=fopen(SPIEGELLOG,"w"))==NULL)goto BADOPENFIL;
	fprintf(SpiegLogF,"\nCOMMAND-LINE OPTIONS\n");
	fprintf(SpiegLogF,"-b: Write Buffer Ring Factor: %d\n",GBufFac);
	fprintf(SpiegLogF,"-c: CD Speed Limit: %d Kb/Sec\n",*CdSpeedP);
	fprintf(SpiegLogF,"-d: DVD Speed Limit: %d Kb/Sec\n",*DvdSpeedP);
	fprintf(SpiegLogF,"-p: Real-Time Priority: %d\n",RtParm.sched_priority);
	return;
NOREALTIME:printf("Real-Time FIFO Scheduling Not Set %s\n",Here);goto STOP;
BADOPENFIL:printf("Failed Open File %s\n",Here);goto STOP;
STOP:printf("errno Says: %s\n",strerror(errno));exit(EXIT_FAILURE);
} /* Handle Command Line Options */

void STARTUP(void) {
	const char *Here="@STARTUP";
	int Kyn;
	char DevAtapi[16];
/* Open Device File */
	printf("DISK NEEDED NOW\n");
	Kyn=GETYN("\tProceed? [%c]\n",'Y');
	if(Kyn=='N')goto QUIT;
	strncpy(DevAtapi,"/dev/",16);
	GETSTR("\tSPECIFY DEVICE IN /dev/ [%s]\n",&DevAtapi[5],DEFLTDEV,11);
	GDevFil=open(DevAtapi,O_RDONLY);
	if(GDevFil==-1)goto BADDEVOPEN;
/* Start Unit */
	memset(&GRequest,0,sizeof(GRequest));
	GRequest.Cmd[0]=0x1B; /* START STOP UNIT */
	GRequest.Cmd[4]=1; /* No-Load Start */
	GRequest.XferDirn=SG_DXFER_NONE;
	if(DOPACKET(GDevFil,"START STOP UNIT",&GRequest)!=0)goto BADSTART;
/* Inquiry Data */
	memset(&GRequest,0,sizeof(GRequest));
	GRequest.Cmd[0]=0x12; /* INQUIRY */
	GRequest.Cmd[4]=96; /* Allocation Length */
	GRequest.XferP=GTransBuf;
	GRequest.XferLen=sizeof(GTransBuf);
	GRequest.XferDirn=SG_DXFER_FROM_DEV;
	if(DOPACKET(GDevFil,"INQUIRY",&GRequest)!=0)goto BADINQ;
	SHOWINQUIRY();
/* Test Unit Ready */
	memset(&GRequest,0,sizeof(GRequest));
	GRequest.Cmd[0]=0x00; /* TEST UNIT READY */
	GRequest.XferDirn=SG_DXFER_NONE;
	if(DOPACKET(GDevFil,"TEST UNIT READY",&GRequest)!=0) goto BADTEST;
	printf("DEVICE FILE OPENED, UNIT STARTED, UNIT TESTED READY\n");
	return;
QUIT:printf("Quit To Insert Media %s\n",Here);goto STOP;
BADDEVOPEN:printf("Open Device File Failed %s\n",Here);goto STOP;
BADSTART:printf("Start Unit Failed %s\n",Here);goto STOP;
BADINQ:printf("Failed Inquiry %s\n",Here);goto STOP;
BADTEST:printf("Failed Test Unit Ready %s\n",Here);goto STOP;
STOP:printf("errno Says: %s\n",strerror(errno));exit(EXIT_FAILURE);
} /* Unit Startup */

uint8_t CONFIGEVAL(void) {
	const char *Here="@CONFIGEVAL";
	int i,j,DescN;
	uint8_t WriteMode,AddLength,Active;
	uint16_t Jshort,CurProfl,ProflN,FeatrN;
/* Display Individual Configuration Features */
	memset(&GRequest,0,sizeof(GRequest));
	GRequest.Cmd[0]=0x46; /* GET CONFIGURATION */
	GRequest.Cmd[1]=2; /* Single Feature Mode */
	Jshort=0;
	BIGEND2(&GRequest.Cmd[2],&Jshort); /* Profile List */
	Jshort=sizeof(GTransBuf);
	BIGEND2(&GRequest.Cmd[7],&Jshort); /* Allocation Length */
	GRequest.XferP=GTransBuf;
	GRequest.XferLen=sizeof(GTransBuf);
	GRequest.XferDirn=SG_DXFER_FROM_DEV;
	if(DOPACKET(GDevFil,"GET CONFIGURATION: Profile List",&GRequest)!=0)goto BADCONFIG;
	fprintf(SpiegLogF,"\nGET CONFIGURATION: Profile List\n");
/* Header Adds Eight Bytes To Profile List */
	BIGEND2(&CurProfl,GTransBuf+6);
	BIGEND2(&FeatrN,GTransBuf+8);
	if(FeatrN!=0)goto PROFLERR;
	printf("Current Profile: %04Xh, %s\n",CurProfl,PROFILES(CurProfl,&WriteMode));
	AddLength=(uint8_t)GTransBuf[11];
	DescN=AddLength>>2;
	fprintf(SpiegLogF,"Profile List Entries: %d, Active [Y] & Inactive [N]\n",DescN);
	for(i=0;i<DescN;i++) {
		j=(i<<2)+12;
		BIGEND2(&ProflN,GTransBuf+j);
		Active=(GTransBuf[j+2]&1)?'Y':'N';
		fprintf(SpiegLogF,"%04Xh [%c] %s\n",ProflN,Active,PROFILES(ProflN,&WriteMode));
	} /* Profile List */
	fprintf(SpiegLogF,"Current Profile: %04Xh, %s\n",CurProfl,PROFILES(CurProfl,&WriteMode));
	if(WriteMode==0xFF)goto WRONGMEDIUM;
	return WriteMode; /* CDR/RW Returns 0x01 */
BADCONFIG:printf("Failed Configuration Request %s\n",Here);goto STOP;
PROFLERR:printf("Wrong Feature Number For Profile List %s\n",Here);goto STOP;
WRONGMEDIUM:printf("Unsuitable Medium Or Drive Identified %s\n",Here);goto STOP;
STOP:printf("errno Says: %s\n",strerror(errno));exit(EXIT_FAILURE);
} /* Hardware Configuration */

const char *PROFILES(uint16_t Code,uint8_t *WriteModeP) {
	*WriteModeP=0xFF; /* Default For Unsuitable Medium */
	switch(Code) {
		case 0x0001:
		return "Non-Removable Re-Writable Disk";
		case 0x0002:
		return "Removable Re-Writable Disk";
		case 0x0003:
		return "Magneto-Optical Sector-Erasable";
		case 0x0004:
		return "Optical Write-Once";
		case 0x0005:
		return "Advance Storage Magneto-Optical";
		case 0x0008:
		return "Read-Only CD-ROM";
		case 0x0009:
		*WriteModeP=0x01;
		return "Write-Once CD-R";
		case 0x000A:
		*WriteModeP=0x01;
		return "Re-Writable CD-RW";
		case 0x0010:
		return "Read-Only DVD-ROM";
		case 0x0011:
		*WriteModeP=0x20;
		return "DVD-R Write-Once Using Sequential Recording";
		case 0x0012:
		return "Re-Writable DVD-RAM";
		case 0x0013:
		return "DVD-RW Re-Writable Using Restricted Over-Write";
		case 0x0014:
		*WriteModeP=0x20;
		return "DVD-RW Re-Writable Using Sequential Recording";
		case 0x0015:
		return "Dual-Layer DVD-R Using Sequential Recording";
		case 0x0016:
		return "Dual-Layer DVD-R Using Layer-Jump Recording";
		case 0x0017:
		return "Dual-Layer DVD-RW Re-Writable";
		case 0x0018:
		return "DVD-Download Disk Recording";
		case 0x001A:
		return "DVD+RW Re-Writable";
		case 0x001B:
		*WriteModeP=0x48; /* Write Parms Unused, Under-Run Protected */
		return "DVD+R Recordable";
		case 0x0020:
		return "Read-Only Double Density DDCD-ROM";
		case 0x0021:
		return "Write-Once Double Density DDCD-R";
		case 0x0022:
		return "Re-Writable Double Density DDCD-RW";
		case 0x002A:
		return "DVD+RW Re-Writable Dual Layer";
		case 0x002B:
		return "DVD+R Recordable Dual Layer";
		case 0x0040:
		return "Blu-Ray BD-ROM";
		case 0x0041:
		return "Blu-Ray BD-R Recordable In Sequential Recording Mode";
		case 0x0042:
		return "Blu-Ray BD-R Recordable In Random Recording Mode";
		case 0x0043:
		return "Blu-Ray BD-RE Re-Writable";
		case 0x0050:
		return "Read-Only HD DVD-ROM";
		case 0x0051:
		return "Write-Once HD DVD-R";
		case 0x0052:
		return "Re-Writable HD DVD-RAM";
		case 0x0053:
		return "Re-Writable HD DVD-RW";
		case 0x0058:
		return "Dual-Layer HD DVD-R Write-Once";
		case 0x005A:
		return "Dual-Layer HD DVD-RW Re-Writable";
		case 0xFFFF:
		return "Non-Standard Profile";
		default:
		return "Reserved Profile Number";
	}
} /* Profile Numbers */
