/* Copyright 2011 Frank Mitchell - Wildebeest License */
/* iface.c - Spiegel v2.0 - Linux Kernel 2.6.16+ */
/* Contact: mitchell@wyatt672earp.force9.co.uk */

#define IFACE
#include "spiegel.h"
#undef IFACE

int GDevFil; /* Device File */
int GBufFac; /* Write Buffer Ring Factor */
off_t GIsoSize; /* Iso File Size Bytes */
uint32_t GDevCache; /* CDRW Hardware Buffer Size */
uint32_t GSeekLBA; /* Last Valid Seek Sector */
ROOTPATH GIsoFilPath; /* UFS Root Path */
SECTOR GTransBuf; /* Global Sector Buffer */
PARMSTOR GWritePage; /* Write Page Spec */
FILE *SpiegLogF; /* Spiegel Log File */
MMCREQ GRequest; /* Linux ATAPI Request */

void SYSEVAL(int argc,char **argv) {
	int J;
	char TestChar;
	struct sched_param RtParm;
	fprintf(SpiegLogF,"\nENVIRONMENT\n");
	J=sizeof(int);
	fprintf(SpiegLogF,"sizeof(int): %d\n",J);
	J=sizeof(off_t);
	fprintf(SpiegLogF,"sizeof(off_t): %d\n",J);
	TestChar=0xFF;
	fprintf(SpiegLogF,"0xFF char Value: %d\n",TestChar);
	fprintf(SpiegLogF,"Program Name: %s\n",argv[0]);
	fprintf(SpiegLogF,"Command Line Arguments: %d\n",argc);
	fprintf(SpiegLogF,"\nREALTIME SCHEDULING PRIORITY\n");
	J=sched_getscheduler(0);
	fprintf(SpiegLogF,"Scheduling Policy: %s\n",SCHEDPOLICY(J));
	J=sched_getparam(0,&RtParm);
	fprintf(SpiegLogF,"Realtime Scheduling Value: %d\n",RtParm.sched_priority);
	return;
} /* Misc System Data */

const char *SCHEDPOLICY(int J) {
	const char *Here="@SCHEDPOLICY";
	switch(J) {
		case SCHED_OTHER:
		printf("WARNING: Non-Root User: No Enhanced Priority\n");
		return "Standard Round-Robin Time-Sharing";
		case SCHED_FIFO:
		return "Pre-Emptive Real-Time First-In First-Out";
		case SCHED_RR:
		return "Real-Time Round-Robin";
		default:
		goto NOPRIORITY;
	} /* Real-Time Categories */
NOPRIORITY:printf("Unknown Scheduling Priority %s\n",Here);goto STOP;
STOP:printf("errno Says: %s\n",strerror(errno));exit(EXIT_FAILURE);
} /* Scheduling Policies */

int DOPACKET(int DevFil,const char *Desc,MMCREQ *RequestP) {
	uint8_t SenseKey,SenseBuf[16];
	sg_io_hdr_t SgReq;
/* FreeBSD Uses struct ata_ioc_request & IOCATAREQUEST ioctl */
	memset(&SgReq,0,sizeof(SgReq));
	SgReq.interface_id='S';
	SgReq.dxfer_direction=RequestP->XferDirn;
	SgReq.cmd_len=12; /* Using ATAPI Commands */
	SgReq.mx_sb_len=sizeof(SenseBuf);
	SgReq.dxfer_len=RequestP->XferLen;
	SgReq.dxferp=RequestP->XferP;
	SgReq.cmdp=RequestP->Cmd;
	SgReq.sbp=SenseBuf;
	if(ioctl(DevFil,SG_IO,&SgReq)<0)return -2;
	if(SgReq.masked_status==0)return 0;
	fprintf(SpiegLogF,"\n%s: Masked Status: %d\n",Desc,SgReq.masked_status);
	fprintf(SpiegLogF,"errno Says: %s\n",strerror(errno));
/* SCSI Sense Codes */
	SenseKey=SenseBuf[2]&0x0F;
	fprintf(SpiegLogF,"SCSI Sense Key: %02Xh, %s\n",SenseKey,SENSEDESCS(SenseKey));
	fprintf(SpiegLogF,"Additional Sense Code ASC: %02Xh\n",SenseBuf[12]);
	fprintf(SpiegLogF,"Sense Code Qualifier ASCQ: %02Xh\n",SenseBuf[13]);
	fflush(NULL); /* Flush All Buffered Output */
	return SgReq.masked_status;
} /* PACKET ioctl For SCSI Command */

const char *SENSEDESCS(uint8_t SenseKey) {
	switch(SenseKey) {
		case 0x00:
		return "No Specific Sense Key Information";
		case 0x01:
		return "Last Command Complete, Recovered From Error";
		case 0x02:
		return "Logical Unit Not Ready, Cannot Be Accessed";
		case 0x03:
		return "Non-Recovered Medium Error";
		case 0x04:
		return "Non-Recoverable Hardware Failure";
		case 0x05:
		return "Illegal Request, Illegal Data Parameter";
		case 0x06:
		return "Unit Attention, Medium Changed Or Target Reset";
		case 0x07:
		return "Data Protected, Read or Write Failed";
		case 0x08:
		return "Encountered Blank Medium, End Of Formatted Data";
		case 0x09:
		return "Vendor Specific Sense Key";
		case 0x0A:
		return "Copy Command Aborted";
		case 0x0B:
		return "Target Aborted Command, Should Retry";
		case 0x0C:
		return "Search Found Equal Comparison";
		case 0x0D:
		return "Data Overflowed End Of Partition";
		case 0x0E:
		return "Source Data Miscompared Against Medium";
		case 0x0F:
		return "Reserved Sense Key";
		default:
		return "See SPC Sense Keys";
	}
} /* Sense Keys */

void SHOWINQUIRY(void) {
	uint8_t Jbyte;
	fprintf(SpiegLogF,"\nINQUIRY\n");
	Jbyte=((uint8_t)GTransBuf[0]&0xE0)>>5;
	fprintf(SpiegLogF,"Peripheral Qualifier: %u, %s\n",Jbyte,PERIPHQUALS(Jbyte));
	Jbyte=(uint8_t)GTransBuf[0]&31;
	fprintf(SpiegLogF,"Device Type: %02Xh, %s\n",Jbyte,DEVTYPES(Jbyte));
	fprintf(SpiegLogF,"Medium Is Removable: %c\n",(GTransBuf[1]&128)?'Y':'N');
	Jbyte=(uint8_t)GTransBuf[2]&7;
	fprintf(SpiegLogF,"ANSI Version: %s\n",ANSIVERSIONS(Jbyte));
	fprintf(SpiegLogF,"Vendor ID String: %1.8s\n",&GTransBuf[8]);
	fprintf(SpiegLogF,"Product ID String: %1.16s\n",&GTransBuf[16]);
	fprintf(SpiegLogF,"Product Revision: %1.4s\n",&GTransBuf[32]);
	return;
} /* INQUIRY Response */

const char *PERIPHQUALS(uint8_t Code) {
	switch(Code) {
		case 0:
		return "Peripheral Connected, Connection Undetermined";
		case 1:
		return "Peripheral Not Connected, Though Is Supported";
		case 2:
		return "Reserved Peripheral Qualifier";
		case 3:
		return "Device Server Cannot Support A Peripheral Here";
		default:
		return "Vendor Specific Peripheral Qualifier";
	}
} /* Peripheral Qualifiers */

const char *DEVTYPES(uint8_t DevType) {
	switch(DevType) {
		case 0x00:
		return "Direct Access Block Device, Magnetic Disc";
		case 0x01:
		return "Sequential Access Device, Magnetic Tape";
		case 0x02:
		return "Printer Device";
		case 0x03:
		return "Processor Device";
		case 0x04:
		return "Write Once Device, Optical Disc";
		case 0x05:
		return "CD/DVD Device";
		case 0x06:
		return "Scanner Device";
		case 0x07:
		return "Optical Memory Device, Optical Disc";
		case 0x08:
		return "Medium Changer Device, Jukebox";
		case 0x09:
		return "Communications Device";
		case 0x0A: case 0x0B:
		return "ASC IT8 Device, Graphic Arts Pre-Press";
		case 0x0C:
		return "Storage Array Controller, RAID";
		case 0x0D:
		return "Enclosure Services Device";
		case 0x0E:
		return "Simplified Direct Access Device";
		case 0x0F:
		return "Optical Card Reader/Writer";
		case 0x10:
		return "Bridge Controller Commands";
		case 0x11:
		return "Object-Based Storage Device";
		case 0x12:
		return "Automation/Drive Interface";
		case 0x13:
		return "Security Manager Device";
		case 0x1E:
		return "Well Known Logical Unit";
		case 0x1F:
		return "Unknown - No Device Type";
		default:
		return "Reserved Device Type";
	}
} /* Peripheral Device Types */

const char *ANSIVERSIONS(uint8_t Code) {
	switch(Code) {
		case 0x00:
		return "ATAPI Device Not Claiming SCSI Conformance";
		case 0x01:
		return "Device Complies To ANSI X3.131:1986 (SCSI-1)";
		case 0x02:
		return "Device Complies To ANSI X3.131:1994 (SCSI-2)";
		case 0x03:
		return "Device Complies To ANSI INCITS 301-1997 (SPC)";
		case 0x04:
		return "Device Complies To ANSI INCITS 351-2001 (SPC-2)";
		case 0x05:
		return "Device Complies To ANSI INCITS 408-2005 (SPC-3)";
		case 0x06:
		return "Device Complies To SPC-4";
		default:
		return "Future ANSI Standard (?)";
	}
} /* ANSI Versions Implemented */

void DEFLTWRITE(void) {
	const char *Here="@DEFLTWRITE";
	uint16_t Jshort;
/* MODE SENSE Default Write Parameters Mode Page */
	memset(&GRequest,0,sizeof(GRequest));
	GRequest.Cmd[0]=0x5A; /* MODE SENSE */
	GRequest.Cmd[1]=8; /* Disable Block Descriptors */
	GRequest.Cmd[2]=0x85; /* Default Write Params */
	Jshort=64; /* Global Write Page Size */
	BIGEND2(&GRequest.Cmd[7],&Jshort);
	GRequest.XferP=GWritePage; /* Global Write Page */
	GRequest.XferLen=64; /* Global Write Page Size */
	GRequest.XferDirn=SG_DXFER_FROM_DEV;
	if(DOPACKET(GDevFil,"MODE SENSE Default Write Parameters",&GRequest)!=0)goto ERRSENSWR;
/* Adjust Write Parameters Page */
	memset(GWritePage,0,8); /* Mode Header Reserved Or Obsolete */
	GWritePage[8]&=0x3F; /* Write Params Page Code */
	GWritePage[10]&=0x6F; /* Ensure Zero Test Write Bit */
/* MODE SELECT Clears Test Track Data */
	memset(&GRequest,0,sizeof(GRequest));
	GRequest.Cmd[0]=0x55; /* MODE SELECT */
	GRequest.Cmd[1]=0x10; /* Standard Page Format */
	Jshort=GWritePage[9]+10; /* Header Plus Total Page Bytes */
	BIGEND2(&GRequest.Cmd[7],&Jshort);
	GRequest.XferP=GWritePage; /* Global Write Page */
	GRequest.XferLen=64; /* Global Write Page Size */
	GRequest.XferDirn=SG_DXFER_TO_DEV;
	if(DOPACKET(GDevFil,"MODE SELECT Write Parameters",&GRequest)!=0)goto ERRSELWR;
	return;
ERRSENSWR:printf("Error Sensing Write Params %s\n",Here);goto STOP;
ERRSELWR:printf("Error Selecting Write Params %s\n",Here);goto STOP;
STOP:printf("errno Says: %s\n",strerror(errno));exit(EXIT_FAILURE);
} /* Reset To Default Write Page */

const char *NOTIFICATIONS(uint8_t Code) {
	switch(Code) {
		case 1:
		return "Operational State Change";
		case 2:
		return "Power Management Event";
		case 3:
		return "External Request";
		case 4:
		return "Media Status Event";
		case 5:
		return "Multiple Initiator Control";
		case 6:
		return "Device Busy Event";
		default:
		return "Reserved Event/Status Notification";
	}
} /* Event Status Notification Types */

void BIGEND2(void *Out16,void *In16) {
	uint8_t *OutByteP,*InByteP;
	OutByteP=(uint8_t*)Out16;
	InByteP=(uint8_t*)In16;
	*OutByteP=*(InByteP+1);
	*(OutByteP+1)=*InByteP;
	return;
} /* Reverse 2 Bytes To Other Endian */

void BIGEND4(void *Out32,void *In32) {
	uint8_t *OutByteP,*InByteP;
	OutByteP=(uint8_t*)Out32;
	InByteP=(uint8_t*)In32;
	*OutByteP=*(InByteP+3);
	*(OutByteP+1)=*(InByteP+2);
	*(OutByteP+2)=*(InByteP+1);
	*(OutByteP+3)=*InByteP;
	return;
} /* Reverse 4 Bytes To Other Endian */

void SPLITEND2(void *Out16,void *HiByte,void *LoByte) {
	uint8_t *OutByteP,*HiByteP,*LoByteP;
	OutByteP=(uint8_t*)Out16;
	HiByteP=(uint8_t*)HiByte;
	LoByteP=(uint8_t*)LoByte;
	*OutByteP=*LoByteP;
	*(OutByteP+1)=*HiByteP;
	return;
} /* Assemble 2 Byte Integer */

int GETYN(const char *Prompt,char DefaultVal) {
	char Buf[80];
	int Clen,Ryn;
	do {
		printf(Prompt,DefaultVal);
		memset(Buf,0,80);
		fgets(Buf,80,stdin);
		if(strcmp(Buf,"\n")==0)return toupper(DefaultVal);
		Clen=strlen(Buf);
		Ryn=toupper(Buf[Clen-2]);
	} while(Ryn!='Y'&&Ryn!='N');
	return Ryn;
} /* Get Y/N Response To Prompt */

int GETINT(const char *Prompt,int DefaultVal) {
	char Buf[80];
	int ScanVal,Jret;
	do {
		printf(Prompt,DefaultVal);
		memset(Buf,0,80);
		fgets(Buf,80,stdin);
		if(strcmp(Buf,"\n")==0)return DefaultVal;
		ScanVal=sscanf(Buf," %d",&Jret);
	} while(ScanVal!=1||ScanVal==EOF);
	return Jret;
} /* Get Integer Response To Prompt */

void GETSTR(const char *Prompt,char *RetStr,const char *Default,size_t ResLen) {
	char *GetStr;
	ssize_t GotLen;
	printf(Prompt,Default);
	memset(RetStr,0,ResLen);
	GetStr=calloc(ResLen,1);
	GotLen=getline(&GetStr,&ResLen,stdin);
	if(GetStr[GotLen-1]=='\n')GotLen--;
	GetStr[GotLen]=0;
	if(GetStr[0]!=0)strncpy(RetStr,GetStr,ResLen);
	else strncpy(RetStr,Default,ResLen);
	free(GetStr);
	return;
} /* Get String Response To Prompt */
