/* Copyright 2011 Frank Mitchell - Wildebeest License */
/* spiegel.h - Spiegel v2.0 - Linux Kernel 2.6.16+ */
/* Contact: mitchell@wyatt672earp.force9.co.uk */

#define _FILE_OFFSET_BITS 64
#define _XOPEN_SOURCE 700

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>

#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sched.h>

#include <sys/stat.h>
#include <sys/ioctl.h>

#include <scsi/sg.h>

/* Version & Licensing Messages */
#define SAPPMES "Spiegel v2.0 CD/DVD Iso Image Writer, Kernel 2.6.16+"
#define CPRTMES "Copyright 2011 Frank Mitchell - Wildebeest License"

#define MINTRACKL 300 /* MMC Minimum Track Length */
#define UFSPATHL 320 /* Max UFS Pathname Bytes & Null */
#define WRITE10MAX 65536 /* Max WRITE(10) Bytes Under FreeBSD v8.0+ */

#define DEFLTDEV "sr0"
#define SPIEGELLOG "./Spiegel.log"

typedef struct {
	int XferDirn; /* Data Transfer Direction */
	char *XferP; /* Data Transfer Buffer */
	int XferLen; /* Data Transfer Length */
	uint8_t Cmd[12]; /* MMC Command*/
} MMCREQ; /* Linux ATAPI Request */

typedef char ROOTPATH[UFSPATHL]; /* UFS Root Path */
typedef char SECTOR[2048]; /* CD/DVD Sector Store */
typedef char PARMSTOR[64]; /* Write Page Storage */

#ifndef IFACE
extern int GDevFil; /* Device File */
extern int GBufFac; /* Write Buffer Ring Factor */
extern off_t GIsoSize; /* Iso File Size Bytes */
extern uint32_t GDevCache; /* CDRW Hardware Buffer Size */
extern uint32_t GSeekLBA; /* Last Valid Seek Sector */
extern ROOTPATH GIsoFilPath; /* UFS Root Path */
extern SECTOR GTransBuf; /* Global Sector Buffer */
extern PARMSTOR GWritePage; /* Write Page Spec */
extern FILE *SpiegLogF; /* Spiegel Log File */
extern MMCREQ GRequest; /* Linux ATAPI Request */
#endif

/* driver.c */
int main(int,char**); /* Driver */
void OPTEVAL(int,char**,uint32_t*,uint32_t*); /* Handle Command Line Options */
void STARTUP(void); /* Unit Startup */
uint8_t CONFIGEVAL(void); /* Hardware Configuration */
const char *PROFILES(uint16_t,uint8_t*); /* Profile Numbers */

/* iface.c */
void SYSEVAL(int,char**); /* Misc System Data */
const char *SCHEDPOLICY(int); /* Scheduling Policies */
int DOPACKET(int,const char*,MMCREQ*); /* PACKET ioctl For SCSI Command */
const char *SENSEDESCS(uint8_t); /* Sense Keys */
void SHOWINQUIRY(void); /* INQUIRY Response */
const char *PERIPHQUALS(uint8_t); /* Peripheral Qualifiers */
const char *DEVTYPES(uint8_t); /* Peripheral Device Types */
const char *ANSIVERSIONS(uint8_t); /* ANSI Versions Implemented */
void DEFLTWRITE(void); /* Reset To Default Write Page */
const char *NOTIFICATIONS(uint8_t); /* Event Status Notification Types */
void BIGEND2(void*,void*); /* Reverse 2 Bytes To Other Endian */
void BIGEND4(void*,void*); /* Reverse 4 Bytes To Other Endian */
void SPLITEND2(void*,void*,void*); /* Assemble 2 Byte Integer */
int GETYN(const char*,char); /* Get Y/N Response To Prompt */
int GETINT(const char*,int); /* Get Integer Response To Prompt */
void GETSTR(const char*,char*,const char*,size_t); /* Get String Response To Prompt */

/* cddev.c */
void DRIVEEVAL(void); /* Hardware Parameter Pages */
void SHOWCAPABILITIES(uint16_t*); /* CD Capabilities & Mechanical Status Page */
void SHOWWRITEPARAMS(const char*); /* Write Parameters Mode Page */
const char *WRITETYPES(uint8_t); /* Write Types */
const char *TRACKMODES(uint8_t); /* Track Modes */
const char *MULTISESSCLOSES(uint8_t); /* Multisession Closures */
const char *DATABLOCKTYPES(uint8_t); /* Data Block Types */
const char *SESSIONFORMATS(uint8_t); /* Session Format Types */
int DISCEVAL(uint16_t*,uint16_t*,uint16_t*); /* Disc Evaluation */
int DISCINFO(uint16_t*,uint16_t*,uint16_t*); /* CDB 0x51, READ DISC INFORMATION */
const char *DISCSTATUSS(uint8_t); /* Disc Statuses */
const char *LASTSESSIONS(uint8_t); /* Last Session States */
const char *DISCTYPES(uint8_t); /* Disc Types */
uint32_t SESSEVAL(uint32_t*,uint16_t,uint16_t); /* Sessions Evaluation */
uint32_t TRACKINFO(uint32_t*); /* CDB 0x52, READ TRACK INFORMATION */
const char *DATAMODES(uint8_t); /* Data Modes */

/* writecyc.c */
uint32_t WRITESPEED(uint32_t,uint32_t,uint32_t,uint8_t,const char*); /* Limit Streaming Speed */
uint8_t WRITEPARMS(uint8_t,uint8_t); /* Set Up Write Parameters Mode Page */
int TESTBURN(uint8_t,uint32_t,uint16_t,uint8_t); /* CDR Writer Driver */
int CLOSEWAIT(uint8_t); /* CLOSE SESSION When Not Test Mode */
