/* Copyright 2011 Frank Mitchell - Wildebeest License */
/* writecyc.c - Spiegel v2.0 - Linux Kernel 2.6.16+ */
/* Contact: mitchell@wyatt672earp.force9.co.uk */

#include "spiegel.h"

uint32_t WRITESPEED(uint32_t WriteStart,uint32_t WriteEnd,
  uint32_t WritePerf,uint8_t StreamCtl,const char *Desc) {
	const char *Here="@WRITESPEED";
	int i,j,k,NumDesc,StartByte;
	uint32_t StartLBA,EndLBA,ReadPerf,StartPerf,EndPerf,Jlong;
	uint16_t Jshort;
	uint8_t TypeField[2]={16,20};
	ReadPerf=0x7FFFFFFF; /* Reset Same */
	Jlong=1000; /* Time Millisec */
	Jshort=28; /* Parameter List Length */
	memset(GTransBuf,0,sizeof(GTransBuf));
	GTransBuf[0]=StreamCtl;
	BIGEND4(&GTransBuf[4],&WriteStart);
	BIGEND4(&GTransBuf[8],&WriteEnd);
	BIGEND4(&GTransBuf[12],&ReadPerf);
	BIGEND4(&GTransBuf[16],&Jlong); /* Read Time */
	BIGEND4(&GTransBuf[20],&WritePerf);
	BIGEND4(&GTransBuf[24],&Jlong); /* Write Time */
/* Set Streaming Speed */
	memset(&GRequest,0,sizeof(GRequest));
	GRequest.Cmd[0]=0xB6; /* SET STREAMING */
	BIGEND2(&GRequest.Cmd[9],&Jshort);
	GRequest.XferP=GTransBuf;
	GRequest.XferLen=sizeof(GTransBuf);
	GRequest.XferDirn=SG_DXFER_TO_DEV;
	if(DOPACKET(GDevFil,"SET STREAMING",&GRequest)!=0)goto ERRSETSPEED;
/* MMC-2 Read & Write Performance Descriptors */
	fprintf(SpiegLogF,"\nSET STREAMING: %s\n",Desc);
	Jshort=120; /* Max Number Descriptors */
	for(i=0;i<2;i++) {
		memset(&GRequest,0,sizeof(GRequest));
		GRequest.Cmd[0]=0xAC; /* GET PERFORMANCE */
		GRequest.Cmd[1]=TypeField[i]; /* Nominal Perfs */
		BIGEND2(&GRequest.Cmd[8],&Jshort);
		GRequest.XferP=GTransBuf;
		GRequest.XferLen=sizeof(GTransBuf);
		GRequest.XferDirn=SG_DXFER_FROM_DEV;
		if(DOPACKET(GDevFil,"GET PERFORMANCE",&GRequest)!=0)goto BADPERFGET;
/* Header Adds Eight Bytes To Performance Descriptor Locations */
		BIGEND4(&Jlong,GTransBuf+0);
		j=Jlong+8;
		memset(&GTransBuf[j],0,2048-j);
		NumDesc=Jlong>>4;
		fprintf(SpiegLogF,"GET PERFORMANCE: %s - Nominal Descriptors: %d\n",
		  (GTransBuf[4]&2)?"Write":"Read",NumDesc);
		for(k=0;k<NumDesc;k++) {
			StartByte=8+(k<<4);
			BIGEND4(&StartLBA,GTransBuf+StartByte);
			BIGEND4(&StartPerf,GTransBuf+StartByte+4);
			BIGEND4(&EndLBA,GTransBuf+StartByte+8);
			BIGEND4(&EndPerf,GTransBuf+StartByte+12);
			fprintf(SpiegLogF,"LBA: %8d - %8d  Speed KB/Sec: %6d - %6d\n",
			  StartLBA,EndLBA,StartPerf,EndPerf);
		} /* List Performance Descriptors */
		if(i==0)ReadPerf=EndPerf;
		else WritePerf=EndPerf;
	} /* Read & Write Descriptors With 10% Tolerance */
	printf("Max Read & Write Speeds KB/Sec Now: %d, %d\n",ReadPerf,WritePerf);
	return WritePerf;
ERRSETSPEED:printf("Error Setting Write Streaming %s\n",Here);goto STOP;
BADPERFGET:printf("Failed Performance Request %s\n",Here);goto STOP;
STOP:printf("errno Says: %s\n",strerror(errno));exit(EXIT_FAILURE);
} /* Limit Streaming Speed */

uint8_t WRITEPARMS(uint8_t WriteMode,uint8_t NextSess) {
	const char *Here="@WRITEPARMS";
	int Kyn;
	char *Desc;
	uint16_t Jshort;
	uint32_t Jlong;
/* Global Write Parameters Mode Page */
	if(WriteMode&0x08) {
		fprintf(SpiegLogF,"\nWrite Parameters Mode Page Not Used With DVD+R\n");
		printf("CAUTION: DVD+R ONLY ALLOWS WRITE RUN: NO TESTS\n");
		return WriteMode;
	} /* DVD+R Ignores Write Parameters */
	WriteMode&=0x2F; /* Restore Plain Write Type */
	Kyn=GETYN("\tWill This Be A Test Run? [%c]\n",'Y');
	if(Kyn=='Y') {
		WriteMode|=0x10; /* Test Run */
		Kyn=GETYN("\tUse Under-Run Protection? [%c]\n",'N');
		if(Kyn=='Y')WriteMode|=0x40; /* Under-Run Protected */
	}
	else WriteMode|=0x40; /* Write Run Protected Anyway */
/* Header Adds Eight Bytes To Mode Page Locations */
	memset(GWritePage,0,8); /* Header Reserved Or Obsolete */
	GWritePage[8]&=0x3F; /* Write Params Page Code */
	GWritePage[10]=WriteMode; /* Write Type, Test Condition */
	GWritePage[11]=NextSess; /* Multi-Session, Uninterrupted Track Mode */
	GWritePage[12]=0x08; /* Mode 1 Data Block Type */
	GWritePage[13]=16; /* Choose 32KB Link Size */
	GWritePage[14]=0; /* Reserved */
	GWritePage[15]=0; /* Host Application Code */
	GWritePage[16]=0; /* Mode 1 Data Disk Session Format */
	GWritePage[17]=0; /* Reserved */
        Jlong=16; /* DVD Fixed Packet Size */
	BIGEND4(GWritePage+18,&Jlong);
	memset(GWritePage+24,0,36); /* Media Catalog, ISRC, Sub-Header */
/* MODE SELECT With Test Write Bit, Write Type & Under-Run Protection */
	memset(&GRequest,0,sizeof(GRequest));
	GRequest.Cmd[0]=0x55; /* MODE SELECT */
	GRequest.Cmd[1]=0x10; /* Standard Page Format */
	Jshort=GWritePage[9]+10; /* Header Plus Total Page Bytes */
	BIGEND2(&GRequest.Cmd[7],&Jshort);
	GRequest.XferP=GWritePage; /* Global Write Page */
	GRequest.XferLen=64; /* Global Write Page Size */
	GRequest.XferDirn=SG_DXFER_TO_DEV;
	if(DOPACKET(GDevFil,"MODE SELECT Write Parameters",&GRequest)!=0)goto ERRSELWR;
/* MODE SENSE Write Parameters Mode Page */
	memset(&GRequest,0,sizeof(GRequest));
	GRequest.Cmd[0]=0x5A; /* MODE SENSE */
	GRequest.Cmd[1]=8; /* Disable Block Descriptors */
	GRequest.Cmd[2]=0x05; /* Current Write Params Mode Page */
	Jshort=sizeof(GTransBuf);
	BIGEND2(&GRequest.Cmd[7],&Jshort);
	GRequest.XferP=GTransBuf;
	GRequest.XferLen=sizeof(GTransBuf);
	GRequest.XferDirn=SG_DXFER_FROM_DEV;
	if(DOPACKET(GDevFil,"MODE SENSE Current Write Parameters",&GRequest)!=0)goto ERRSENSWR;
	if(WriteMode&0x10)Desc="Test Run Settings";
	else Desc="Write Run Settings";
	SHOWWRITEPARAMS(Desc);
	return WriteMode;
ERRSENSWR:printf("Error Sensing Write Params %s\n",Here);goto STOP;
ERRSELWR:printf("Error Selecting Write Params %s\n",Here);goto STOP;
STOP:printf("errno Says: %s\n",strerror(errno));exit(EXIT_FAILURE);
} /* Set Up Write Parameters Mode Page */

int TESTBURN(uint8_t WriteMode,uint32_t WriteStart,uint16_t SessTrackN,uint8_t CloseFunc) {
	const char *Here="@TESTBURN";
	int RunFail,IsoFildes,WriteK,BufJ,BufRingK,RingBufTot;
	int32_t Jlong,BufSectors,StartSect,BlocksToDo;
	int16_t Jshort;
	off_t *OffSetP;
	void **RingPP;
	if(WriteMode&0x10)fprintf(SpiegLogF,"\nSPIEGEL TEST WRITE\n");
	else fprintf(SpiegLogF,"\nSPIEGEL WRITE RUN\n");
	RunFail=0; /* Write Fail Flag */
	BufSectors=(int32_t)WRITE10MAX>>11;
/* Total Ring Buffer Bytes = Hardware Cache Size /8 *GBufFac */
	RingBufTot=(GDevCache>>3)*GBufFac;
	BufRingK=RingBufTot/(size_t)WRITE10MAX;
	fprintf(SpiegLogF,"Ring Buffer Count: %u\n",BufRingK);
	OffSetP=calloc(BufRingK,sizeof(off_t));
	if(OffSetP==NULL)goto BADCALLOC;
	RingPP=calloc(BufRingK,sizeof(void*));
	if(RingPP==NULL)goto BADCALLOC;
	fflush(NULL); /* Flush All Buffered Output */
	if((IsoFildes=open(GIsoFilPath,O_RDONLY))==-1)goto BADOPENFIL;
	for(BufJ=0;BufJ<BufRingK;BufJ++) {
		RingPP[BufJ]=calloc((size_t)WRITE10MAX,1);
		if(RingPP[BufJ]==NULL)goto BADCALLOC;
		OffSetP[BufJ]=(off_t)WRITE10MAX*BufJ;
		pread(IsoFildes,RingPP[BufJ],(size_t)WRITE10MAX,OffSetP[BufJ]);
	} /* Memory Buffer Ring */
	printf("WAIT UNTIL CLOSE SESSION MESSAGE APPEARS\n");
	BlocksToDo=(GIsoSize+2047)>>11; /* Partial Block? */
	WriteK=0; /* Buffer Serial Count */
	BufJ=0; /* Burn Cycle Index Mod BufRingK */
/* Move To Last Valid Seek Sector */
	memset(&GRequest,0,sizeof(GRequest));
	GRequest.Cmd[0]=0x2B; /* SEEK(10) */
	BIGEND4(&GRequest.Cmd[2],&GSeekLBA); /* Last Seek LBA */
	GRequest.XferDirn=SG_DXFER_NONE;
	if(DOPACKET(GDevFil,"SEEK(10)",&GRequest)!=0)goto SEEKFAIL;
	StartSect=WriteStart;
	while(1) {
		Jshort=(BlocksToDo<BufSectors)?BlocksToDo:BufSectors; /* # Blocks */
		Jlong=(int32_t)Jshort<<11; /* # Bytes + Padding */
		memset(&GRequest,0,sizeof(GRequest));
		GRequest.Cmd[0]=0x2A; /* WRITE(10) */
		BIGEND4(&GRequest.Cmd[2],&StartSect); /* LBA Start */
		BIGEND2(&GRequest.Cmd[7],&Jshort); /* # Blocks */
		GRequest.XferP=RingPP[BufJ];
		GRequest.XferLen=Jlong; /* # Bytes */
		GRequest.XferDirn=SG_DXFER_TO_DEV;
		if(DOPACKET(GDevFil,"WRITE(10)",&GRequest)!=0)goto WRITEFAIL;
		WriteK++;
		StartSect+=Jshort;
		BlocksToDo-=Jshort;
		if(BlocksToDo<=0)goto SYNCHCACHE;
		OffSetP[BufJ]+=RingBufTot;
		pread(IsoFildes,RingPP[BufJ],(size_t)WRITE10MAX,OffSetP[BufJ]);
		BufJ+=(BufJ==(BufRingK-1))?-(BufRingK-1):1; /* Burn Cycle Mod BufRingK */
	} /* Loop FreeBSD ioctl WRITE(10) 64K */
WRITEFAIL: /* Flag Run Failure */
	RunFail=1;
	printf("\nRUN FAILED: WRITE COMMAND ERROR\n");
	fprintf(SpiegLogF,"RUN FAILED: WRITE COMMAND ERROR\n");
	fprintf(SpiegLogF,"errno Says: %s\n",strerror(errno));
SYNCHCACHE: /* Synchronize Cache */
	if(BlocksToDo!=0)fprintf(SpiegLogF,"ERROR! Blocks Discrepancy: %d\n",BlocksToDo);
	fprintf(SpiegLogF,"Buffer Serial Count: %d\n",WriteK);
	fprintf(SpiegLogF,"Final Buffer Block Count: %d\n",Jshort);
	memset(&GRequest,0,sizeof(GRequest));
	GRequest.Cmd[0]=0x35; /* SYNCHRONIZE CACHE */
	BIGEND4(&GRequest.Cmd[2],&StartSect); /* Start LBA */
	Jshort=0; /* Range Includes All Remaining Sectors */
	BIGEND2(&GRequest.Cmd[7],&Jshort);
	GRequest.XferDirn=SG_DXFER_NONE;
	if(DOPACKET(GDevFil,"SYNCHRONIZE CACHE",&GRequest)!=0) {
		printf("Failed SYNCHRONIZE CACHE\n");
		fprintf(SpiegLogF,"Failed SYNCHRONIZE CACHE\n");
	} /* SYNCHRONIZE CACHE Messages */
/* Close Track */
	memset(&GRequest,0,sizeof(GRequest));
	GRequest.Cmd[0]=0x5B; /* CLOSE TRACK/SESSION */
	GRequest.Cmd[2]=1; /* Close Track */
	Jshort=SessTrackN; /* Close Incomplete Track */
	BIGEND2(&GRequest.Cmd[4],&Jshort);
	GRequest.XferDirn=SG_DXFER_NONE;
	if(DOPACKET(GDevFil,"CLOSE TRACK",&GRequest)!=0) {
		printf("Failed CLOSE TRACK\n");
		fprintf(SpiegLogF,"Failed CLOSE TRACK\n");
	} /* CLOSE TRACK Messages */
	if(WriteMode&0x10) {
		printf("Skip CLOSE SESSION In Test Mode\n");
	} /* Test Mode */
	else RunFail=CLOSEWAIT(CloseFunc);
	close(IsoFildes);
	for(BufJ=0;BufJ<BufRingK;BufJ++) {
		free(RingPP[BufJ]);
	} /* Memory Buffer Ring */
	free(RingPP);
	free(OffSetP);
	fprintf(SpiegLogF,"\n");
	return RunFail;
BADCALLOC:printf("Couldn't Allocate Dynamic Memory %s\n",Here);goto STOP;
BADOPENFIL:printf("Couldn't open() Iso File %s\n",Here);goto STOP;
SEEKFAIL:printf("Failed SEEK(10) To Last Valid Seek Sector %s\n",Here);goto STOP;
STOP:printf("errno Says: %s\n",strerror(errno));exit(EXIT_FAILURE);
} /* CDR Writer Driver */

int CLOSEWAIT(uint8_t CloseFunc) {
	const char *Here="@CLOSEWAIT";
	int j=0,RunFail=0;
	uint8_t Jbyte;
	uint16_t Jshort;
	float Time;
/* Close Session */
	printf("WAIT DURING CLOSE SESSION\n");
	memset(&GRequest,0,sizeof(GRequest));
	GRequest.Cmd[0]=0x5B; /* CLOSE TRACK/SESSION */
	GRequest.Cmd[1]=1; /* Immed Set */
	GRequest.Cmd[2]=CloseFunc; /* Close Session */
	GRequest.XferDirn=SG_DXFER_NONE;
	if(DOPACKET(GDevFil,"CLOSE SESSION",&GRequest)!=0) {
		RunFail=1;
		printf("Failed CLOSE SESSION\n");
		fprintf(SpiegLogF,"Failed CLOSE SESSION\n");
	} /* CLOSE SESSION Messages */
/* Get Event/Status Notification */
	fprintf(SpiegLogF,"\nGET EVENT/STATUS NOTIFICATION Device Busy Class\n");
	memset(&GRequest,0,sizeof(GRequest));
	GRequest.Cmd[0]=0x4A; /* GET EVENT/STATUS NOTIFICATION */
	GRequest.Cmd[1]=1; /* Immed/Polled */
	GRequest.Cmd[4]=0x40; /* Device Busy Class */
	Jshort=sizeof(GTransBuf);
	BIGEND2(&GRequest.Cmd[7],&Jshort);
	GRequest.XferP=GTransBuf;
	GRequest.XferLen=sizeof(GTransBuf);
	GRequest.XferDirn=SG_DXFER_FROM_DEV;
	if(DOPACKET(GDevFil,"GET EVENT/STATUS NOTIFICATION",&GRequest)!=0)goto BADNOTIFGET;
	Jbyte=GTransBuf[2]&0x07;
	fprintf(SpiegLogF,"Notification Returned: %u, %s\n",Jbyte,NOTIFICATIONS(Jbyte));
	BIGEND2(&Jshort,GTransBuf+6);
	Time=0.1*(float)Jshort;
	printf("Predicted Time Remaining: %5.1f Sec\n",Time);
	j=GTransBuf[5]; /* Device Busy Status */
	while(j!=0) {
		if(usleep(500000)<0)goto BADSLEEP; /* Microsec Sleep */
		if(DOPACKET(GDevFil,"GET EVENT/STATUS NOTIFICATION",&GRequest)!=0)goto BADNOTIFGET;
		j=GTransBuf[5]; /* Device Busy Status */
	} /* Silent Polling */
	if(RunFail==0)printf("Completed CLOSE SESSION\n"); /* Delayed While Polling */
	return RunFail;
BADNOTIFGET:printf("Failed Event/Status Notification Request %s\n",Here);goto STOP;
BADSLEEP:printf("Failed usleep() %s\n",Here);goto STOP;
STOP:printf("errno Says: %s\n",strerror(errno));exit(EXIT_FAILURE);
} /* CLOSE SESSION When Not Test Mode */
